import os
from typing import TYPE_CHECKING

from PySide6.QtCore import Qt, Signal
from PySide6.QtGui import QAction, QKeySequence, QResizeEvent, QShowEvent
from PySide6.QtWidgets import (
    QAbstractItemView,
    QAbstractScrollArea,
    QFrame,
    QHBoxLayout,
    QMenu,
    QTabBar,
    QTreeView,
    QWidget,
)
from typing_extensions import override

from conan_explorer import conan_version
from conan_explorer.app.base_ui.loading import LoaderGui
from conan_explorer.conan_wrapper.types import ConanPkg, ConanRef, pretty_print_pkg_info
from conan_explorer.ui.common.model import re_register_signal
from conan_explorer.ui.plugin import PluginDescription, PluginInterfaceV1

from .file_controller import PackageFileExplorerController
from .sel_controller import MultiPkgSelectionMode, PackageSelectionController
from .sel_model import PkgSelectionType

if TYPE_CHECKING:
    from conan_explorer.ui.fluent_window import FluentWindow
    from conan_explorer.ui.main_window import BaseSignals


class LocalConanPackageExplorer(PluginInterfaceV1):
    conan_pkg_selected = Signal(str, ConanPkg, PkgSelectionType)
    # cut works globally
    cut_files_reset = Signal()  # for when e.g. a copy is requested or a new cut

    def __init__(
        self,
        parent: QWidget,
        plugin_description: PluginDescription,
        base_signals: "BaseSignals",
        page_widgets: "FluentWindow.PageStore",
    ):
        super().__init__(parent, plugin_description, base_signals, page_widgets)
        from .package_explorer_ui import Ui_Form

        self._ui = Ui_Form()
        self._ui.setupUi(self)
        self.load_signal.connect(self.load)
        self._is_expanded = False

    def load(self):
        assert self._base_signals
        assert self._page_widgets
        self._pkg_sel_ctrl = PackageSelectionController(
            self,
            self._ui.package_select_view,
            self._ui.package_filter_edit,
            self.conan_pkg_selected,
            self._base_signals,
            self._page_widgets,
        )
        self._pkg_tabs_ctrl = [
            PackageFileExplorerController(
                self,
                self._ui.package_file_view,
                self._ui.package_path_label,
                self.cut_files_reset,
                self._base_signals,
                self._page_widgets,
            )
        ]
        self._file_cntx_menu = None
        self.set_themed_icon(self._ui.refresh_button, "icons/refresh.svg")
        self.set_themed_icon(self._ui.show_sizes_button, "icons/bar_chart.svg")
        self.set_themed_icon(self._ui.expand_all_button, "icons/expand_all.svg")

        # connect pkg selection controller
        self._ui.package_select_view.header().setSortIndicator(0, Qt.SortOrder.AscendingOrder)
        self._ui.package_select_view.setContextMenuPolicy(
            Qt.ContextMenuPolicy.CustomContextMenu
        )
        self._ui.package_select_view.customContextMenuRequested.connect(
            self.on_selection_context_menu_requested
        )
        self._init_selection_context_menu()
        self._ui.refresh_button.clicked.connect(self._pkg_sel_ctrl.on_pkg_refresh_clicked)
        self._ui.show_sizes_button.clicked.connect(self.on_show_sizes)
        self._ui.package_filter_edit.textChanged.connect(self._pkg_sel_ctrl.set_filter_wildcard)
        self._ui.expand_all_button.clicked.connect(self.on_expand_all)

        self.conan_pkg_selected.connect(self.on_pkg_selection_change)
        self._ui.package_path_label.setText("<Package path>")
        self._ui.package_path_label.setAlignment(
            Qt.AlignmentFlag.AlignRight
        )  # must be called after every text set
        self._ui.package_tab_widget.currentChanged.connect(self.on_tab_index_changed)
        # removes X of new Tab button
        self._ui.package_tab_widget.tabBar().setTabButton(
            1,
            QTabBar.ButtonPosition.RightSide,
            None,  # type: ignore
        )
        self._ui.package_tab_widget.tabCloseRequested.connect(self.on_close_tab)
        self._ui.package_tab_widget.tabBar().setSelectionBehaviorOnRemove(
            QTabBar.SelectionBehavior.SelectLeftTab
        )
        self.updateGeometry()
        self.resize_filter()

    # Display Events

    def reload_themed_icons(self):
        super().reload_themed_icons()
        self._init_selection_context_menu()
        self._init_pkg_file_context_menu()

    @override
    def showEvent(self, a0: QShowEvent) -> None:
        # only update the first time
        self._pkg_sel_ctrl.refresh_pkg_selection_view(force_update=False)
        return super().showEvent(a0)

    @override
    def resizeEvent(self, a0: QResizeEvent) -> None:
        for pkg_file_exp_ctrl in self._pkg_tabs_ctrl:
            pkg_file_exp_ctrl.resize_file_columns()
        super().resizeEvent(a0)

    # Tab control

    def tab_close_requested(self, pkg_ctrl_to_close):
        self.on_close_tab(self._pkg_tabs_ctrl.index(pkg_ctrl_to_close))

    def on_close_tab(self, index: int):
        self._ui.package_tab_widget.removeTab(index)
        self._pkg_tabs_ctrl.pop(index)

    def on_tab_index_changed(self, index: int):
        assert self._base_signals
        assert self._page_widgets

        # adds new tab
        if self._ui.package_tab_widget.count() == index + 1:
            tab = QWidget(self._ui.package_tab_widget)
            horizontal_layout = QHBoxLayout(tab)
            horizontal_layout.setSpacing(0)
            horizontal_layout.setObjectName("horizontalLayout")
            horizontal_layout.setContentsMargins(0, 0, 0, 0)
            tab.setLayout(horizontal_layout)
            file_explorer_view = QTreeView(tab)
            horizontal_layout.addWidget(file_explorer_view)
            file_explorer_view.setFrameShape(QFrame.Shape.NoFrame)
            file_explorer_view.setSizeAdjustPolicy(
                QAbstractScrollArea.SizeAdjustPolicy.AdjustToContents
            )
            file_explorer_view.setEditTriggers(
                QAbstractItemView.EditTrigger.EditKeyPressed
                | QAbstractItemView.EditTrigger.SelectedClicked
            )
            file_explorer_view.setTabKeyNavigation(True)
            file_explorer_view.setDragDropMode(QAbstractItemView.DragDropMode.DragDrop)
            file_explorer_view.setDefaultDropAction(Qt.DropAction.TargetMoveAction)
            file_explorer_view.setIndentation(15)
            file_explorer_view.setUniformRowHeights(True)
            file_explorer_view.setItemsExpandable(True)
            file_explorer_view.setSortingEnabled(True)
            file_explorer_view.setAnimated(True)
            file_explorer_view.setSelectionMode(
                QAbstractItemView.SelectionMode.ExtendedSelection
            )

            self._pkg_tabs_ctrl.append(
                PackageFileExplorerController(
                    self,
                    file_explorer_view,
                    self._ui.package_path_label,
                    self.cut_files_reset,
                    self._base_signals,
                    self._page_widgets,
                )
            )

            self._ui.package_tab_widget.insertTab(index, tab, "New tab")
            self._ui.package_tab_widget.setCurrentIndex(index)
        else:
            ctrl = self._pkg_tabs_ctrl[self._ui.package_tab_widget.currentIndex()]
            if ctrl._model:
                self._ui.package_path_label.setText(ctrl._model.rootPath())

    def on_pkg_selection_change(self, conan_ref: str, pkg: ConanPkg, type: PkgSelectionType):
        # init/update the context menu
        for ctrl in self._pkg_tabs_ctrl:
            re_register_signal(
                ctrl._view.customContextMenuRequested, self.on_file_context_menu_requested
            )
        self._init_pkg_file_context_menu()
        cfr = ConanRef.loads(conan_ref)
        disp_ref = f"{cfr.name}/{cfr.version}"  # only package/version
        if type == PkgSelectionType.export:
            disp_ref += " (export)"
        idx = self._ui.package_tab_widget.currentIndex()
        self._ui.package_tab_widget.setTabText(idx, disp_ref)
        self._ui.package_tab_widget.setTabToolTip(idx, pretty_print_pkg_info(pkg))
        self._ui.package_tab_widget.tabBar().setExpanding(True)
        ctrl = self._pkg_tabs_ctrl[idx]

        ctrl.on_pkg_selection_change(conan_ref, pkg, type)
        if ctrl._model:
            self._ui.package_path_label.setText(ctrl._model.rootPath())

    # Selection view general buttons and functions

    def resize_filter(self):
        # resize filter splitter to roughly match file view splitter
        sizes = self._ui.splitter.sizes()
        offset = self._ui.package_filter_label.width() + self._ui.refresh_button.width()
        self._ui.splitter_filter.setSizes([sizes[0] - offset - 5, sizes[1] + 5])

    def resize_for_show_sizes(self):
        sizes = self._ui.splitter.sizes()
        sizes_sum = sum(sizes)
        self._ui.splitter.setSizes([int(sizes_sum / 2), int(sizes_sum / 2)])

    def on_show_sizes(self):
        self._pkg_sel_ctrl.on_show_sizes()
        self.resize_for_show_sizes()

    def on_expand_all(self):
        if self._is_expanded:
            self._ui.package_select_view.collapseAll()
            self.set_themed_icon(self._ui.expand_all_button, "icons/expand_all.svg")
            self._is_expanded = False
            return

        loader = LoaderGui(self)
        loader.load(
            self,
            self._ui.package_select_view.expandAll,
        )
        self.set_themed_icon(self._ui.expand_all_button, "icons/collapse_all.svg")
        self._is_expanded = True

    # Selection view context menu

    def _init_selection_context_menu(self):
        self.select_cntx_menu = QMenu()

        self.copy_ref_action = QAction("Copy reference", self)
        self.set_themed_icon(self.copy_ref_action, "icons/copy_link.svg")
        self.select_cntx_menu.addAction(self.copy_ref_action)
        self.copy_ref_action.triggered.connect(self._pkg_sel_ctrl.on_copy_ref_requested)

        self.show_conanfile_action = QAction("Show conanfile", self)
        self.set_themed_icon(self.show_conanfile_action, "icons/file_preview.svg")
        self.select_cntx_menu.addAction(self.show_conanfile_action)
        self.show_conanfile_action.triggered.connect(
            self._pkg_sel_ctrl.on_show_conanfile_requested
        )

        self.install_ref_action = QAction("Re(install) package", self)
        self.set_themed_icon(self.install_ref_action, "icons/download_pkg.svg")
        self.select_cntx_menu.addAction(self.install_ref_action)
        self.install_ref_action.triggered.connect(self._pkg_sel_ctrl.on_install_ref_requested)

        # always have show_build_info_action so access to it is possible in ConanV2
        # even if it does nothing
        self.show_build_info_action = QAction("Show package build info", self)
        if conan_version.major != 2:  # Currently not doable
            self.set_themed_icon(self.show_build_info_action, "icons/download.svg")
            self.select_cntx_menu.addAction(self.show_build_info_action)
            self.show_build_info_action.triggered.connect(self._pkg_sel_ctrl.on_show_build_info)

        self.diff_pkg_action = QAction("Compare packages", self)
        self.set_themed_icon(self.diff_pkg_action, "icons/difference.svg")
        self.select_cntx_menu.addAction(self.diff_pkg_action)
        self.diff_pkg_action.triggered.connect(self._pkg_sel_ctrl.on_diff_requested)

        self.remove_ref_action = QAction("Remove package(s)", self)
        self.set_themed_icon(self.remove_ref_action, "icons/delete.svg")
        self.select_cntx_menu.addAction(self.remove_ref_action)
        self.remove_ref_action.triggered.connect(self._pkg_sel_ctrl.on_remove_ref_requested)

    def on_selection_context_menu_requested(self, position):
        # no multiselect
        sources = self._pkg_sel_ctrl.get_selected_pkg_source_items()
        selection_mode = self._pkg_sel_ctrl.get_selection_mode(sources)
        if selection_mode == MultiPkgSelectionMode.single_ref:
            self._context_menu_set_all(True)
            self.diff_pkg_action.setVisible(False)
            self.show_build_info_action.setVisible(False)
        elif selection_mode == MultiPkgSelectionMode.multi_ref:
            self._context_menu_set_all(False)
            self.copy_ref_action.setVisible(True)
            self.remove_ref_action.setVisible(True)
        elif selection_mode == MultiPkgSelectionMode.single_pkg_or_export:
            self._context_menu_set_all(True)
            self.diff_pkg_action.setVisible(False)
        elif selection_mode == MultiPkgSelectionMode.multi_pkg:
            self._context_menu_set_all(False)
            self.diff_pkg_action.setVisible(True)
            self.remove_ref_action.setVisible(True)
        elif selection_mode == MultiPkgSelectionMode.single_editable:
            self._context_menu_set_all(False)
            self.copy_ref_action.setVisible(True)
        else:  # invalid
            self._context_menu_set_all(False)
            self.copy_ref_action.setVisible(True)  # always works
            self.remove_ref_action.setVisible(True)  # TODO

        self.select_cntx_menu.exec(self._ui.package_select_view.mapToGlobal(position))

    def _context_menu_set_all(self, enabled: bool):
        self.copy_ref_action.setVisible(enabled)
        self.show_build_info_action.setVisible(enabled)
        self.show_conanfile_action.setVisible(enabled)
        self.install_ref_action.setVisible(enabled)
        self.remove_ref_action.setVisible(enabled)
        self.diff_pkg_action.setVisible(enabled)

    # Package File Explorer context menu

    def _init_pkg_file_context_menu(self):
        if self._file_cntx_menu:
            return
        # for pkg_file_exp_ctrl in self._pkg_tabs_ctrl:
        self._file_cntx_menu = QMenu()
        self._open_fm_action = QAction("Show in File Manager", self)
        self.set_themed_icon(self._open_fm_action, "icons/file_explorer.svg")
        self._file_cntx_menu.addAction(self._open_fm_action)
        self._open_fm_action.triggered.connect(self.on_open_file_in_file_manager)

        self._copy_as_path_action = QAction("Copy as Path", self)
        self.set_themed_icon(self._copy_as_path_action, "icons/copy_to_clipboard.svg")
        self._file_cntx_menu.addAction(self._copy_as_path_action)
        self._copy_as_path_action.triggered.connect(self.on_copy_file_as_path)

        self._edit_file_action = QAction("Edit file", self)
        self.set_themed_icon(self._edit_file_action, "icons/edit_file.svg")
        self._file_cntx_menu.addAction(self._edit_file_action)
        self._edit_file_action.triggered.connect(self.on_edit_file)

        self._open_terminal_action = QAction("Open terminal here", self)
        self.set_themed_icon(self._open_terminal_action, "icons/cmd.svg")
        self._file_cntx_menu.addAction(self._open_terminal_action)
        self._open_terminal_action.triggered.connect(self.on_open_terminal_in_dir)

        self._file_cntx_menu.addSeparator()

        self._rename_action = QAction("Rename", self)
        self.set_themed_icon(self._rename_action, "icons/rename.svg")
        self._rename_action.setShortcut(QKeySequence("F2"))
        self._rename_action.setShortcutContext(Qt.ShortcutContext.WidgetWithChildrenShortcut)
        self._file_cntx_menu.addAction(self._rename_action)
        # for the shortcut to work, the action has to be added to a higher level widget
        self.addAction(self._rename_action)
        self._rename_action.triggered.connect(self.on_item_rename)

        self._copy_action = QAction("Copy", self)
        self.set_themed_icon(self._copy_action, "icons/copy.svg")
        self._copy_action.setShortcut(QKeySequence("Ctrl+c"))
        self._copy_action.setShortcutContext(Qt.ShortcutContext.WidgetWithChildrenShortcut)
        self._file_cntx_menu.addAction(self._copy_action)
        # for the shortcut to work, the action has to be added to a higher level widget
        self.addAction(self._copy_action)
        self._copy_action.triggered.connect(self.on_file_copy)

        self._cut_action = QAction("Cut", self)
        self.set_themed_icon(self._cut_action, "icons/cut.svg")
        self._cut_action.setShortcut(QKeySequence("Ctrl+x"))
        self._cut_action.setShortcutContext(Qt.ShortcutContext.WidgetWithChildrenShortcut)
        self._file_cntx_menu.addAction(self._cut_action)
        # for the shortcut to work, the action has to be added to a higher level widget
        self.addAction(self._cut_action)
        self._cut_action.triggered.connect(self.on_file_cut)

        self._paste_action = QAction("Paste", self)
        self.set_themed_icon(self._paste_action, "icons/paste.svg")
        self._paste_action.setShortcut(QKeySequence("Ctrl+v"))  # Qt.CTRL + Qt.Key_V))
        self._paste_action.setShortcutContext(Qt.ShortcutContext.WidgetWithChildrenShortcut)
        self.addAction(self._paste_action)
        self._file_cntx_menu.addAction(self._paste_action)
        self._paste_action.triggered.connect(self.on_file_paste)

        self._new_folder_action = QAction("New folder", self)
        self.set_themed_icon(self._new_folder_action, "icons/opened_folder.svg")
        self.addAction(self._new_folder_action)
        self._file_cntx_menu.addAction(self._new_folder_action)
        self._new_folder_action.triggered.connect(self.on_new_folder)

        self._delete_action = QAction("Delete", self)
        self.set_themed_icon(self._delete_action, "icons/delete.svg")
        self._delete_action.setShortcut(QKeySequence(Qt.Key.Key_Delete))
        self._file_cntx_menu.addAction(self._delete_action)
        self._delete_action.setShortcutContext(Qt.ShortcutContext.WidgetWithChildrenShortcut)
        self.addAction(self._delete_action)
        self._delete_action.triggered.connect(self.on_file_delete)

        self._file_cntx_menu.addSeparator()

        self._add_link_action = QAction("Add link to App Grid", self)
        self.set_themed_icon(self._add_link_action, "icons/add_link.svg")
        self._file_cntx_menu.addAction(self._add_link_action)
        self._add_link_action.triggered.connect(self.on_add_app_link_from_file)

    def on_file_context_menu_requested(self, position):
        """Disable some context menu items depending on context"""
        if not self._file_cntx_menu:
            return
        self._add_link_action.setVisible(True)
        self._edit_file_action.setVisible(True)
        self._rename_action.setVisible(True)

        tab_idx = self._ui.package_tab_widget.currentIndex()
        # Add Link only works on actual packages
        pkg_type = self._pkg_tabs_ctrl[tab_idx].get_conan_pkg_type()
        if pkg_type == PkgSelectionType.export:
            self._add_link_action.setVisible(False)
        else:
            self._add_link_action.setVisible(True)

        # multiselect options
        paths = self._pkg_tabs_ctrl[tab_idx].get_selected_pkg_paths()
        if len(paths) > 1:  # no multiselect
            self._add_link_action.setVisible(False)
            self._edit_file_action.setVisible(False)
            self._rename_action.setVisible(False)
            self._open_fm_action.setVisible(False)
            self._copy_as_path_action.setVisible(False)
            self._open_terminal_action.setVisible(False)
        elif os.path.isdir(paths[0]):
            self._add_link_action.setVisible(False)
            self._edit_file_action.setVisible(False)
            self._rename_action.setVisible(True)
            self._open_fm_action.setVisible(True)
            self._copy_as_path_action.setVisible(True)
            self._open_terminal_action.setVisible(True)

        self._file_cntx_menu.exec(self._ui.package_file_view.mapToGlobal(position))

    def select_local_package_from_ref(self, conan_ref: str):
        return self._pkg_sel_ctrl.select_local_package_from_ref(conan_ref)

    def on_new_folder(self, model_index):
        return self._pkg_tabs_ctrl[self._ui.package_tab_widget.currentIndex()].on_new_folder(
            model_index
        )

    def on_open_file_in_file_manager(self, model_index):
        return self._pkg_tabs_ctrl[
            self._ui.package_tab_widget.currentIndex()
        ].on_open_file_in_file_manager(model_index)

    def on_copy_file_as_path(self, model_index):
        return self._pkg_tabs_ctrl[
            self._ui.package_tab_widget.currentIndex()
        ].on_copy_file_as_path()

    def on_edit_file(self, model_index):
        return self._pkg_tabs_ctrl[self._ui.package_tab_widget.currentIndex()].on_edit_file()

    def on_open_terminal_in_dir(self, model_index):
        return self._pkg_tabs_ctrl[
            self._ui.package_tab_widget.currentIndex()
        ].on_open_terminal_in_dir()

    def on_item_rename(self, model_index):
        return self._pkg_tabs_ctrl[self._ui.package_tab_widget.currentIndex()].on_item_rename()

    def on_file_copy(self, model_index):
        return self._pkg_tabs_ctrl[self._ui.package_tab_widget.currentIndex()].on_files_copy()

    def on_file_cut(self, model_index):
        return self._pkg_tabs_ctrl[self._ui.package_tab_widget.currentIndex()].on_files_cut()

    def on_file_paste(self, model_index):
        return self._pkg_tabs_ctrl[self._ui.package_tab_widget.currentIndex()].on_files_paste()

    def on_file_delete(self, model_index):
        return self._pkg_tabs_ctrl[self._ui.package_tab_widget.currentIndex()].on_file_delete()

    def on_add_app_link_from_file(self, model_index):
        return self._pkg_tabs_ctrl[
            self._ui.package_tab_widget.currentIndex()
        ].on_add_app_link_from_file()



"""Contains the class for a clickable Qt label"""

from pathlib import Path
from typing import Optional

from PySide6.QtCore import QSize, Qt
from PySide6.QtGui import QColor, QIcon, QPixmap
from PySide6.QtWidgets import QGraphicsColorizeEffect, QGraphicsPixmapItem, QPushButton

from conan_explorer import ICON_SIZE, INVALID_PATH


class ClickableIcon(QPushButton):
    """
    Qt QPushButton with greyable icon, which can react on a mouse click.
    Has advanced icon handling for displaying the best matching icon.
    """

    # TODO flags
    def __init__(self, parent, image=Path(INVALID_PATH), icon_size=ICON_SIZE):
        super().__init__(parent=parent)

        self._greyed_out = True  # Must be ungreyed, when available
        self._grey_effect = QGraphicsColorizeEffect()
        self._grey_effect.setColor(QColor(128, 128, 128, 128))
        self.setGraphicsEffect(self._grey_effect)

        self.set_icon_from_file(image)
        self.setFlat(True)
        self.setIconSize(QSize(icon_size, icon_size))
        self.grey_icon()

    def ungrey_icon(self):
        self._greyed_out = False
        self._grey_effect.setEnabled(False)

    def grey_icon(self):
        self._greyed_out = True
        self._grey_effect.setEnabled(True)

    def set_icon(self, icon: QIcon):
        # convert biggest image to QPixmap and scale
        sizes = icon.availableSizes()
        if len(sizes) > 0:
            icon_pixmap = icon.pixmap(sizes[-1])
            # embedded icons sometimes not scale up correctly, but only increase the canvas behind the icon,
            #  which looks like crap. With opaqueArea.boundingRect we can query the size of
            #  the non-opaque area - the real image
            pixmap = QGraphicsPixmapItem(icon_pixmap)
            image_rect = pixmap.opaqueArea().boundingRect()
            top_left = image_rect.topLeft()
            # copy the smaller image (crop) - start from 0, this adds some margin
            icon_pixmap = icon_pixmap.copy(
                0,
                0,
                int(image_rect.width() + top_left.x()),
                int(image_rect.height() + top_left.y()),
            )
            self._ic = QIcon(icon_pixmap)
        else:
            self._ic = icon
        self.setIcon(self._ic)

    def set_icon_from_file(self, image: Optional[Path]):
        if not image or not image.exists():
            return
        self._image = image
        if self._image.suffix == ".ico":
            self._ic = QIcon(str(self._image))
            self.setIcon(self._ic)
        else:
            pixmap = QPixmap(str(self._image)).toImage()
            icon = QPixmap.fromImage(pixmap).scaled(
                ICON_SIZE, ICON_SIZE, mode=Qt.TransformationMode.SmoothTransformation
            )
            self._ic = QIcon(icon)
            self.setIcon(self._ic)



from pathlib import Path
from typing import TYPE_CHECKING, List, Type, TypeVar

from PySide6.QtCore import Qt, Signal
from PySide6.QtGui import QAction, QIcon
from PySide6.QtWidgets import (
    QFileDialog,
    QInputDialog,
    QMenu,
    QMessageBox,
    QTabWidget,
    QVBoxLayout,
)

import conan_explorer.app as app
from conan_explorer import AUTHOR, BUILT_IN_PLUGIN, user_save_path
from conan_explorer.app.logger import Logger
from conan_explorer.conan_wrapper.types import ConanPkgRef, ConanRef

# using global module pattern
from conan_explorer.settings import AUTO_INSTALL_QUICKLAUNCH_REFS, LAST_CONFIG_FILE

# using global module pattern
from conan_explorer.ui.common import get_themed_asset_icon
from conan_explorer.ui.fluent_window import FluentWindow
from conan_explorer.ui.plugin import PluginInterfaceV1
from conan_explorer.ui.plugin.types import PluginDescription
from conan_explorer.ui.widgets import AnimatedToggle

from .config import UiAppLinkConfig, UiTabConfig
from .model import UiAppLinkModel, UiTabModel
from .tab import TabList

if TYPE_CHECKING:
    from conan_explorer.ui.main_window import BaseSignals
    from conan_explorer.ui.views.app_grid.model import UiAppGridModel


class AppGridView(PluginInterfaceV1):
    load_signal = Signal()  # type: ignore

    def __init__(
        self,
        parent,
        model: "UiAppGridModel",
        base_signals: "BaseSignals",
        page_widgets: FluentWindow.PageStore,
    ):
        plugin_descr = PluginDescription(
            "Conan Quicklaunch", BUILT_IN_PLUGIN, AUTHOR, "", "", "", " ", False, ""
        )
        super().__init__(parent, plugin_descr, base_signals, page_widgets)

        self.tab_widget = QTabWidget(self)
        self.tab_widget.setContentsMargins(0, 0, 0, 0)
        self.tab_widget.setElideMode(Qt.TextElideMode.ElideLeft)
        self.tab_widget.setUsesScrollButtons(True)
        self.setLayout(QVBoxLayout(self))
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().addWidget(self.tab_widget)

        self.model = model
        base_signals.conan_pkg_installed.connect(self.update_conan_info)

        self.tab_widget.tabBar().setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.tab_widget.tabBar().setContentsMargins(0, 0, 0, 0)
        self.tab_widget.tabBar().customContextMenuRequested.connect(
            self.on_tab_context_menu_requested
        )

        self.tab_widget.setMovable(True)
        self.tab_widget.tabBar().tabMoved.connect(self.on_tab_move)
        if self.tab_widget.count() > 0:  # remove the default tab
            self.tab_widget.removeTab(0)
        self.load_signal.connect(self.load)

    def load(self, offset=0):
        """Creates new layout"""
        self._init_right_menu()
        for tab_config in self.model.tabs:
            # need to save object locally, otherwise it can be destroyed in the underlying C++ layer
            tab = TabList(parent=self.tab_widget, model=tab_config)
            self.tab_widget.addTab(tab, tab_config.name)
            tab.load(offset)

        # always show the first tab first
        self.tab_widget.setCurrentIndex(0)

    def _init_right_menu(self):
        # Right Settings menu
        assert self._page_widgets
        quicklaunch_submenu = self._page_widgets.get_side_menu_by_type(type(self))
        assert quicklaunch_submenu
        quicklaunch_submenu.reset_widgets()
        quicklaunch_submenu.add_button_menu_entry(
            "Open Layout File", self.open_config_file_dialog, "icons/opened_folder.svg"
        )
        quicklaunch_submenu.add_button_menu_entry(
            "Add AppLink", self.on_add_link, "icons/add_link.svg"
        )
        quicklaunch_submenu.add_button_menu_entry(
            "Reorder AppLinks", self.on_reorder, "icons/rearrange.svg"
        )
        quicklaunch_submenu.add_menu_line()

        quicklaunch_submenu.add_toggle_menu_entry(
            "Auto-install quicklaunch packages",
            self.on_toggle_auto_install,
            app.active_settings.get_bool(AUTO_INSTALL_QUICKLAUNCH_REFS),
        )

    def reload_themed_icons(self):
        self.re_init(self.model)

    T = TypeVar("T")

    def findChildren(self, type: Type[T]) -> List[T]:
        return super().findChildren(type)  # type: ignore

    def re_init(self, model: "UiAppGridModel", offset=0):
        """To be called, when a new config file is loaded"""
        self.model = model
        # clear all tabs and flag them for later deletion
        self.tab_widget.clear()
        for tab in self.get_tabs():
            tab.deleteLater()
        self.load(offset)

    def re_init_all_app_links(self, force=False):
        for tab in self.get_tabs():
            tab.redraw(force)

    def open_new_app_link_dialog(self):
        # call tab on_app_link_add
        current_tab: TabList = self.tab_widget.widget(self.tab_widget.currentIndex())  # type: ignore
        current_tab.open_app_link_add_dialog()

    def on_tab_move(self):
        """Refresh backend info when tabs are reordered"""
        reordered_tabs = []
        for i in range(self.tab_widget.count()):
            tab: TabList = self.tab_widget.widget(i)  # type: ignore
            reordered_tabs.append(tab.model)
        self.model.tabs = reordered_tabs
        self.model.save()

    def on_tab_context_menu_requested(self, position):
        index = self.tab_widget.tabBar().tabAt(position)
        menu = QMenu()
        self.menu = menu

        rename_action = QAction("Rename", self)
        rename_action.setIcon(QIcon(get_themed_asset_icon("icons/rename.svg")))
        menu.addAction(rename_action)
        rename_action.triggered.connect(lambda: self.on_tab_rename(index))

        remove_action = QAction("Remove", self)
        remove_action.setIcon(QIcon(get_themed_asset_icon("icons/delete.svg")))
        menu.addAction(remove_action)
        remove_action.triggered.connect(lambda: self.on_tab_remove(index))

        new_tab_action = QAction("Add new tab", self)
        new_tab_action.setIcon(QIcon(get_themed_asset_icon("icons/plus.svg")))
        menu.addAction(new_tab_action)
        new_tab_action.triggered.connect(self.on_new_tab)

        menu.exec(self.tab_widget.tabBar().mapToGlobal(position))
        return self.menu  # for testing

    def on_new_tab(self):
        # call tab on_app_link_add
        new_tab_dialog = QInputDialog(self)
        text, accepted = new_tab_dialog.getText(self, "Add tab", "Enter name:")
        if accepted:
            # do nothing on empty text
            if not text:
                return
            # update model
            tab_model = UiTabModel().load(
                UiTabConfig(text, apps=[UiAppLinkConfig()]), self.model
            )
            self.model.tabs.append(tab_model)
            self.model.save()
            # add tab in ui
            tab = TabList(self.tab_widget, model=tab_model)
            tab.load()
            self.tab_widget.addTab(tab, text)

    def on_tab_rename(self, index):
        tab: TabList = self.tab_widget.widget(index)  # type: ignore

        rename_tab_dialog = QInputDialog(self)
        text, accepted = rename_tab_dialog.getText(
            self, "Rename tab", "Enter new name:", text=tab.model.name
        )
        if accepted:
            tab.model.name = text
            self.tab_widget.setTabText(index, text)
            tab.model.save()

    def on_tab_remove(self, index):
        # last tab can't be deleted! # TODO dialog
        if len(self.model.tabs) == 1:
            return
        msg = QMessageBox(parent=self)
        msg.setWindowTitle("Delete tab")
        msg.setText("Are you sure, you want to delete this tab?\t")
        msg.setStandardButtons(
            QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.Cancel
        )
        msg.setIcon(QMessageBox.Icon.Question)
        reply = msg.exec()
        if reply == QMessageBox.StandardButton.Yes:
            self.tab_widget.removeTab(index)
            self.model.tabs.remove(self.model.tabs[index])
            self.model.save()

    def get_tabs(self) -> List[TabList]:
        return self.findChildren(TabList)

    def on_toggle_auto_install(self):
        sender_toggle: AnimatedToggle = self.sender()  # type: ignore
        status = sender_toggle.isChecked()
        app.active_settings.set(AUTO_INSTALL_QUICKLAUNCH_REFS, status)
        # model loads incrementally
        self.model.parent.loadf(app.active_settings.get_string(LAST_CONFIG_FILE))
        sender_toggle.wait_for_anim_finish()

        self.re_init(self.model)

    def open_config_file_dialog(self):
        """ " Open File Dialog and load config file"""
        dialog_path = user_save_path
        config_file_path = Path(app.active_settings.get_string(LAST_CONFIG_FILE))
        if config_file_path.exists():
            dialog_path = config_file_path.parent
        dialog = QFileDialog(
            parent=self,
            caption="Select JSON Config File",
            directory=str(dialog_path),
            filter="JSON files (*.json)",
        )
        dialog.setFileMode(QFileDialog.FileMode.ExistingFile)
        if dialog.exec() == QFileDialog.DialogCode.Accepted:
            new_file = dialog.selectedFiles()[0]
            app.active_settings.set(LAST_CONFIG_FILE, new_file)
            # model loads incrementally
            self.model.parent.loadf(new_file)
            # conan works, model can be loaded
            self.re_init(self.model)  # loads tabs

    def on_add_link(self):
        tab: TabList = self.tab_widget.currentWidget()  # type: ignore
        tab.open_app_link_add_dialog()

    def on_reorder(self):
        tab: TabList = self.tab_widget.currentWidget()  # type: ignore
        tab.app_links[0].on_move()

    def open_new_app_dialog_from_extern(self, app_config: UiAppLinkConfig):
        """Called from pacakge explorer, where tab is unknown"""
        dialog = QInputDialog(self)
        tab_list = list(item.name for item in self.model.tabs)
        model = UiAppLinkModel()
        dialog.setLabelText("Choose a tab for the new AppLink!")
        dialog.setComboBoxItems(tab_list)
        dialog.setWindowTitle("New AppLink")
        if dialog.exec() == QInputDialog.DialogCode.Accepted:
            answer = dialog.textValue()
            for tab in self.get_tabs():
                if answer == tab.model.name:
                    tab.open_app_link_add_dialog(model.load(app_config, tab.model))
                    break

    def update_conan_info(self, conan_ref: str, pkg_id: str):
        if not self.isEnabled():  # the gui is about to shut down
            return
        try:
            for tab in self.get_tabs():
                for app_link in tab.app_links:
                    # Catch shutdown
                    if not self.isEnabled():
                        return
                    if not conan_ref:  # unspecific reload
                        app_link.model.load_from_cache()

                    if app_link.model.conan_ref == conan_ref:
                        # reverse lookup - don't update an icon with other options
                        pkg_info = app.conan_api.get_local_pkg_from_id(
                            ConanPkgRef.loads(conan_ref + ":" + pkg_id)
                        )
                        if (
                            app_link.model.conan_options
                        ):  # only compare options, if user explicitly set them
                            # user options should be a subset of full pkg options
                            if (
                                not app_link.model.conan_options.items()
                                <= pkg_info.get("options", {}).items()
                            ):
                                continue
                        if pkg_id:
                            app_link.model.set_package_folder(
                                app.conan_api.get_package_folder(
                                    ConanRef.loads(conan_ref), pkg_id
                                )
                            )
                        else:
                            app_link.model.load_from_cache()
                    app_link.apply_conan_info()

        except Exception as e:
            Logger().error(f"Can't update AppGrid with conan info {str(e)}")



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'conan_install.ui'
##
## Created by: Qt User Interface Compiler version 6.7.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QAbstractButton, QAbstractItemView, QApplication, QComboBox,
    QDialog, QDialogButtonBox, QFormLayout, QFrame,
    QGridLayout, QHBoxLayout, QHeaderView, QLabel,
    QPushButton, QSizePolicy, QTreeWidget, QTreeWidgetItem,
    QVBoxLayout, QWidget)

from conan_explorer.ui.widgets import AnimatedToggle
from conan_explorer.ui.widgets.conan_line_edit import ConanRefLineEdit

class Ui_Dialog(object):
    def setupUi(self, Dialog):
        if not Dialog.objectName():
            Dialog.setObjectName(u"Dialog")
        Dialog.resize(406, 343)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(Dialog.sizePolicy().hasHeightForWidth())
        Dialog.setSizePolicy(sizePolicy)
        Dialog.setMinimumSize(QSize(400, 100))
        Dialog.setSizeIncrement(QSize(0, 0))
        Dialog.setModal(False)
        self.gridLayout = QGridLayout(Dialog)
        self.gridLayout.setSpacing(2)
        self.gridLayout.setObjectName(u"gridLayout")
        self.gridLayout.setContentsMargins(2, 2, 2, 2)
        self.main_frame = QFrame(Dialog)
        self.main_frame.setObjectName(u"main_frame")
        self.main_frame.setFrameShape(QFrame.StyledPanel)
        self.main_frame.setFrameShadow(QFrame.Raised)
        self.formLayout = QFormLayout(self.main_frame)
        self.formLayout.setObjectName(u"formLayout")
        self.conan_ref_line_edit = ConanRefLineEdit(self.main_frame)
        self.conan_ref_line_edit.setObjectName(u"conan_ref_line_edit")
        sizePolicy.setHeightForWidth(self.conan_ref_line_edit.sizePolicy().hasHeightForWidth())
        self.conan_ref_line_edit.setSizePolicy(sizePolicy)
        self.conan_ref_line_edit.setMinimumSize(QSize(0, 0))
        self.conan_ref_line_edit.setMaximumSize(QSize(1024, 16777215))
        self.conan_ref_line_edit.setMaxLength(1024)
        self.conan_ref_line_edit.setFrame(True)

        self.formLayout.setWidget(0, QFormLayout.FieldRole, self.conan_ref_line_edit)

        self.profile_label = QLabel(self.main_frame)
        self.profile_label.setObjectName(u"profile_label")

        self.formLayout.setWidget(1, QFormLayout.LabelRole, self.profile_label)

        self.profile_cbox = QComboBox(self.main_frame)
        self.profile_cbox.setObjectName(u"profile_cbox")

        self.formLayout.setWidget(1, QFormLayout.FieldRole, self.profile_cbox)

        self.set_default_install_profile_button = QPushButton(self.main_frame)
        self.set_default_install_profile_button.setObjectName(u"set_default_install_profile_button")

        self.formLayout.setWidget(2, QFormLayout.FieldRole, self.set_default_install_profile_button)

        self.conan_opts_label = QLabel(self.main_frame)
        self.conan_opts_label.setObjectName(u"conan_opts_label")
        sizePolicy1 = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.MinimumExpanding)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.conan_opts_label.sizePolicy().hasHeightForWidth())
        self.conan_opts_label.setSizePolicy(sizePolicy1)

        self.formLayout.setWidget(3, QFormLayout.LabelRole, self.conan_opts_label)

        self.options_widget = QTreeWidget(self.main_frame)
        self.options_widget.setObjectName(u"options_widget")
        self.options_widget.setEditTriggers(QAbstractItemView.DoubleClicked|QAbstractItemView.EditKeyPressed)
        self.options_widget.setSortingEnabled(True)
        self.options_widget.setWordWrap(True)

        self.formLayout.setWidget(3, QFormLayout.FieldRole, self.options_widget)

        self.label = QLabel(self.main_frame)
        self.label.setObjectName(u"label")

        self.formLayout.setWidget(0, QFormLayout.LabelRole, self.label)


        self.gridLayout.addWidget(self.main_frame, 0, 0, 1, 1)

        self.bottom_frame = QFrame(Dialog)
        self.bottom_frame.setObjectName(u"bottom_frame")
        self.bottom_frame.setFrameShape(QFrame.StyledPanel)
        self.bottom_frame.setFrameShadow(QFrame.Raised)
        self.verticalLayout = QVBoxLayout(self.bottom_frame)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(2, 2, 2, 2)
        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.update_check_box = AnimatedToggle(self.bottom_frame)
        self.update_check_box.setObjectName(u"update_check_box")
        sizePolicy2 = QSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Fixed)
        sizePolicy2.setHorizontalStretch(0)
        sizePolicy2.setVerticalStretch(0)
        sizePolicy2.setHeightForWidth(self.update_check_box.sizePolicy().hasHeightForWidth())
        self.update_check_box.setSizePolicy(sizePolicy2)

        self.horizontalLayout.addWidget(self.update_check_box)

        self.update_label = QLabel(self.bottom_frame)
        self.update_label.setObjectName(u"update_label")
        self.update_label.setLayoutDirection(Qt.RightToLeft)

        self.horizontalLayout.addWidget(self.update_label)


        self.verticalLayout.addLayout(self.horizontalLayout)

        self.line = QFrame(self.bottom_frame)
        self.line.setObjectName(u"line")
        self.line.setFrameShape(QFrame.Shape.HLine)
        self.line.setFrameShadow(QFrame.Shadow.Sunken)

        self.verticalLayout.addWidget(self.line)

        self.horizontalLayout_2 = QHBoxLayout()
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.auto_install_check_box = AnimatedToggle(self.bottom_frame)
        self.auto_install_check_box.setObjectName(u"auto_install_check_box")
        sizePolicy2.setHeightForWidth(self.auto_install_check_box.sizePolicy().hasHeightForWidth())
        self.auto_install_check_box.setSizePolicy(sizePolicy2)

        self.horizontalLayout_2.addWidget(self.auto_install_check_box)

        self.auto_install_label = QLabel(self.bottom_frame)
        self.auto_install_label.setObjectName(u"auto_install_label")

        self.horizontalLayout_2.addWidget(self.auto_install_label)


        self.verticalLayout.addLayout(self.horizontalLayout_2)

        self.button_box = QDialogButtonBox(self.bottom_frame)
        self.button_box.setObjectName(u"button_box")
        sizePolicy.setHeightForWidth(self.button_box.sizePolicy().hasHeightForWidth())
        self.button_box.setSizePolicy(sizePolicy)
        self.button_box.setOrientation(Qt.Horizontal)
        self.button_box.setStandardButtons(QDialogButtonBox.Cancel|QDialogButtonBox.Ok)
        self.button_box.setCenterButtons(False)

        self.verticalLayout.addWidget(self.button_box)


        self.gridLayout.addWidget(self.bottom_frame, 1, 0, 1, 1)


        self.retranslateUi(Dialog)
        self.button_box.accepted.connect(Dialog.accept)
        self.button_box.rejected.connect(Dialog.reject)

        QMetaObject.connectSlotsByName(Dialog)
    # setupUi

    def retranslateUi(self, Dialog):
        Dialog.setWindowTitle(QCoreApplication.translate("Dialog", u"Conan Install", None))
        self.conan_ref_line_edit.setText("")
        self.profile_label.setText(QCoreApplication.translate("Dialog", u"Profile", None))
        self.set_default_install_profile_button.setText(QCoreApplication.translate("Dialog", u"Set this as default install profile", None))
        self.conan_opts_label.setText(QCoreApplication.translate("Dialog", u"<html><head/><body><p>Options</p></body></html>", None))
        ___qtreewidgetitem = self.options_widget.headerItem()
        ___qtreewidgetitem.setText(1, QCoreApplication.translate("Dialog", u"Value", None));
        ___qtreewidgetitem.setText(0, QCoreApplication.translate("Dialog", u"Name", None));
        self.label.setText(QCoreApplication.translate("Dialog", u"Reference", None))
        self.update_check_box.setText("")
        self.update_label.setText(QCoreApplication.translate("Dialog", u"Update (-u)", None))
        self.auto_install_check_box.setText("")
        self.auto_install_label.setText(QCoreApplication.translate("Dialog", u"Automatically determine best matching package", None))
    # retranslateUi




import xml.dom.minidom as dom
from pathlib import Path

from PySide6.QtCore import QFileInfo
from PySide6.QtGui import QIcon, QImage
from PySide6.QtWidgets import QFileIconProvider

from conan_explorer import INVALID_PATH
from conan_explorer.app.logger import Logger


def get_inverted_asset_image(image_path: Path) -> Path:
    """Inverts a given image and saves it beside the original one with _inv in the name.
    To be used for icons to switch between light and dark mode themes."""
    inverted_img_path = image_path.parent / (
        (image_path.with_suffix("").name + "_inv") + image_path.suffix
    )

    if not inverted_img_path.exists():
        img = QImage(str(image_path))
        img.invertPixels()
        img.save(str(inverted_img_path))
    return inverted_img_path


def get_icon_from_image_file(image_path: Path) -> QIcon:
    return QIcon(str(image_path))


def draw_svg_with_color(svg_path: Path, color="white") -> Path:
    """
    Sets an svg in the desired color for a QtWidget.
    :param color: the disired color as a string in html compatible name
    :param shadow: draws a drop shadow
    """
    if not svg_path or not svg_path.exists():
        Logger().error("Cannot draw invalid SVG file: %s", repr(svg_path))
        return Path(INVALID_PATH)

    # read svg as xml and get the drawing
    with open(svg_path, "r", encoding="utf-8") as svg:
        svg_content = "".join(svg.readlines())
        svg_content = svg_content.replace("\t", "")
    svg_dom = dom.parseString("".join(svg_content))
    svg_paths = svg_dom.getElementsByTagName("path")
    # set color in the dom element
    for path in svg_paths:
        path.setAttribute("fill", color)
    # also replace possible css
    xml_text: str = svg_dom.toxml()
    xml_text = xml_text.replace("#000000", "white")
    # create temporary svg and read into pyqt svg graphics object
    new_svg_path = svg_path.parent / Path(svg_path.stem + "_" + color + svg_path.suffix)
    with open(new_svg_path, "w+", encoding="utf-8") as new_svg:
        new_svg.write(xml_text)
    return new_svg_path


def extract_icon(file_path: Path) -> QIcon:
    """
    Extract icons from a file and save them to specified dir.
    There must be a main qt application created, otherwise the call will crash!
    """

    if file_path.is_file():
        icon_provider = QFileIconProvider()
        file_info = QFileInfo(str(file_path))
        icon = icon_provider.icon(file_info)
        return icon
    else:
        Logger().debug("File %s for icon extraction does not exist.", str(file_path))
    return QIcon()


def get_platform_icon(profile_name: str) -> QIcon:
    """Return an Icon based on the profile name."""
    from conan_explorer.ui.common.theming import get_themed_asset_icon

    profile_name = profile_name.lower()
    if "win" in profile_name:  # I hope people have no random win"s" in their profilename
        return QIcon(get_themed_asset_icon("icons/global/windows.svg"))
    elif "linux" in profile_name:
        return QIcon(get_themed_asset_icon("icons/global/linux.svg"))
    elif "android" in profile_name:
        return QIcon(get_themed_asset_icon("icons/global/android.svg"))
    elif "macos" in profile_name:
        return QIcon(get_themed_asset_icon("icons/global/mac_os.svg"))
    else:
        return QIcon(get_themed_asset_icon("icons/default_pkg.svg"))



"""
This module contains all interfaces and backend functions
containing the applications core functionality.
Settings need to be already set up for usage.
"""

from conan_explorer import Version, conan_version

from .conan_cache import ConanInfoCache
from .conan_worker import ConanWorker
from .unified_api import ConanCommonUnifiedApi

__all__ = ["conan_version", "Version", "ConanCommonUnifiedApi", "ConanInfoCache", "ConanWorker"]


def ConanApiFactory() -> ConanCommonUnifiedApi:
    """Isntantiate ConanApi in the correct version"""
    if conan_version.major == 1:
        from .conanV1 import ConanApi

        return ConanApi()
    elif conan_version.major == 2:
        from .conanV2 import ConanApi

        return ConanApi()
    else:
        raise RuntimeError("Can't recognize Conan version")



from pathlib import Path
from typing import Optional

from PySide6.QtCore import SignalInstance
from PySide6.QtWidgets import QDialog, QFileDialog, QMessageBox, QWidget

import conan_explorer.app as app  # using global module pattern
from conan_explorer.app.logger import Logger
from conan_explorer.conan_wrapper.types import ConanRef
from conan_explorer.ui.common import measure_font_width
from conan_explorer.ui.common.theming import get_themed_asset_icon
from conan_explorer.ui.dialogs import ConanInstallDialog
from conan_explorer.ui.views.app_grid.model import UiAppLinkModel


class AppEditDialog(QDialog):
    def __init__(
        self,
        model: UiAppLinkModel,
        parent: Optional[QWidget],
        pkg_installed_signal: Optional[SignalInstance] = None,
    ):
        super().__init__(parent=parent)
        self._model = model
        self._pkg_installed_signal = pkg_installed_signal
        from .app_edit_dialog_ui import Ui_Dialog

        # without baseinstance, dialog would further needed to be configured
        self._ui = Ui_Dialog()
        self._ui.setupUi(self)

        self.setModal(True)
        self.setWindowTitle("Edit App Link")
        self.setWindowIcon(get_themed_asset_icon("icons/edit.svg", True))

        self._ui.executable_browse_button.setEnabled(False)
        self._ui.icon_browse_button.setEnabled(False)

        self._ui.conan_ref_line_edit.set_loading_callback(self.loading_started_info)
        self._ui.conan_ref_line_edit.completion_finished.connect(self.loading_finished_info)
        self._ui.conan_ref_line_edit.textChanged.connect(self.toggle_browse_buttons)

        # fill up current info
        self._ui.name_line_edit.setText(self._model.name)
        self._ui.conan_ref_line_edit.setText(self._model.conan_ref)
        self._ui.execpath_line_edit.setText(self._model.executable)
        self._ui.is_console_app_checkbox.setChecked(self._model.is_console_application)
        self._ui.icon_line_edit.setText(self._model.icon)
        self._ui.args_line_edit.setText(self._model.args)
        width_to_set = measure_font_width(self._ui.args_line_edit.text()) + 10
        if width_to_set > self.parentWidget().geometry().width():
            width_to_set = self.parentWidget().geometry().width()
        if width_to_set < self.parentWidget().geometry().width() / 2:
            width_to_set = int(self.parentWidget().geometry().width() / 2)
        self.setMinimumWidth(width_to_set)
        conan_options_text = ""
        for option in self._model.conan_options:
            conan_options_text += f"{option}={self._model.conan_options.get(option)}\n"
        self._ui.conan_opts_text_edit.setText(conan_options_text)

        self._ui.install_button.clicked.connect(self.on_install_clicked)
        self._ui.executable_browse_button.clicked.connect(self.on_executable_browse_clicked)
        self._ui.icon_browse_button.clicked.connect(self.on_icon_browse_clicked)

        # enable/disable button, if package_folder is set
        if self._pkg_installed_signal:
            self._pkg_installed_signal.connect(self.toggle_browse_buttons)

        # for some reason OK is not connected at default
        self._ui.button_box.accepted.connect(self.save_data)
        self._ui.button_box.accepted.connect(self.cleanup)
        self._ui.button_box.rejected.connect(self.cleanup)

    def cleanup(self) -> None:
        self._ui.conan_ref_line_edit.cleanup()

    def toggle_browse_buttons(self):
        if not self._ui.conan_ref_line_edit.is_valid:
            self._ui.executable_browse_button.setEnabled(False)
            self._ui.icon_browse_button.setEnabled(False)
            return
        self._ui.executable_browse_button.setEnabled(True)
        self._ui.icon_browse_button.setEnabled(True)

    def on_install_clicked(self):
        dialog = ConanInstallDialog(
            self,
            self._ui.conan_ref_line_edit.text(),
            self._pkg_installed_signal,
            lock_reference=True,
        )
        dialog.show()

    def on_executable_browse_clicked(self):
        _, temp_package_path = app.conan_api.get_best_matching_local_package_path(
            ConanRef.loads(self._ui.conan_ref_line_edit.text()), self.resolve_conan_options()
        )
        if not temp_package_path.exists():  # default path
            temp_package_path = Path.home()
        dialog = QFileDialog(
            parent=self, caption="Select file to start", directory=str(temp_package_path)
        )
        # filter="Images (*.ico *.svg *.jpg)")
        dialog.setFileMode(QFileDialog.FileMode.ExistingFile)
        if dialog.exec() == QFileDialog.DialogCode.Accepted:
            exe_path = Path(dialog.selectedFiles()[0])
            try:
                exe_rel_path = exe_path.relative_to(temp_package_path)
            except Exception:
                msg = QMessageBox(parent=self)
                msg.setWindowTitle("Invalid selection")
                msg.setText(
                    f"The entered path {str(exe_path)} is not in the selected conan package folder!"
                )
                msg.setStandardButtons(QMessageBox.StandardButton.Ok)
                msg.setIcon(QMessageBox.Icon.Critical)
                msg.exec()
                return False
            # use as_posix to always get forward slashes in the relpath
            self._ui.execpath_line_edit.setText(exe_rel_path.as_posix())
            return True

    def on_icon_browse_clicked(self):
        _, temp_package_path = app.conan_api.get_best_matching_local_package_path(
            ConanRef.loads(self._ui.conan_ref_line_edit.text()), self.resolve_conan_options()
        )
        if not temp_package_path.exists():  # default path
            temp_package_path = Path.home()
        dialog = QFileDialog(
            parent=self,
            caption="Select file for icon display",
            directory=str(temp_package_path),
            filter="Images (*.ico *.png *.svg *.jpg)",
        )
        dialog.setFileMode(QFileDialog.FileMode.ExistingFile)
        if dialog.exec() == QFileDialog.DialogCode.Accepted:
            icon_path = Path(dialog.selectedFiles()[0])
            try:
                icon_rel_path = icon_path.relative_to(temp_package_path)
                self._ui.icon_line_edit.setText(str(icon_rel_path))
            except Exception:
                # errors, if it does not resolve
                self._ui.icon_line_edit.setText(str(icon_path))

    def loading_started_info(self):
        self._ui.conan_ref_label.setText("Conan Reference (query in progress)")

    def loading_finished_info(self):
        self._ui.conan_ref_label.setText("Conan Reference (query finished)")

    def resolve_conan_options(self):
        conan_options_text = self._ui.conan_opts_text_edit.toPlainText().splitlines()
        conan_options = {}
        for line in conan_options_text:
            split_values = line.split("=")
            if len(split_values) == 2:
                conan_options.update({split_values[0]: split_values[1]})
            else:
                Logger().warning(f"Wrong format in option: {line}")
        return conan_options

    def save_data(self):
        # check all input validations
        if not self._ui.conan_ref_line_edit.is_valid:
            msg = QMessageBox(parent=self)
            msg.setWindowTitle("Invalid Conan Reference")
            msg.setText("The entered Conan reference has an invalid format!")
            msg.setStandardButtons(QMessageBox.StandardButton.Ok)
            msg.setIcon(QMessageBox.Icon.Critical)
            msg.exec()
            return
        # write back app info
        self._model.name = self._ui.name_line_edit.text()
        self._model.conan_ref = self._ui.conan_ref_line_edit.text()
        self._model.executable = self._ui.execpath_line_edit.text()
        self._model.is_console_application = self._ui.is_console_app_checkbox.isChecked()
        self._model.icon = self._ui.icon_line_edit.text()
        self._model.args = self._ui.args_line_edit.text()
        self._model.conan_options = self.resolve_conan_options()
        self._model.save()
        self.accept()



from datetime import datetime, timedelta

from PySide6.QtCore import (  # type: ignore
    Property,
    QEasingCurve,
    QPoint,
    QPointF,
    QPropertyAnimation,
    QRectF,
    QSequentialAnimationGroup,
    Qt,
    Slot,
)
from PySide6.QtGui import QBrush, QColor, QPainter, QPaintEvent, QPen
from PySide6.QtWidgets import QApplication, QCheckBox
from typing_extensions import override


class AnimatedToggle(QCheckBox):
    ANIM_DURATION_MS = 400
    FIXED_WIDTH = 60
    FIXED_HEIGHT = 50
    THUMB_REL_SIZE = 0.15
    TRACK_REL_SIZE = 0.4

    def __init__(
        self,
        parent=None,
        bar_color=Qt.GlobalColor.gray,
        checked_color="#00B0FF",
        thumb_color=Qt.GlobalColor.white,
        pulse_unchecked_color="#44999999",
        pulse_checked_color="#4400B0EE",
    ):
        super().__init__(parent)
        self._transparent_pen = QPen(Qt.GlobalColor.transparent)
        self._light_grey_pen = QPen(Qt.GlobalColor.lightGray)

        # needed for paintEvent.
        self._background_color_brush = QBrush(bar_color)
        self._bar_checked_brush = QBrush(QColor(checked_color).lighter())
        self._disabled_checked_brush = QBrush(QColor(bar_color).darker())
        self._thumb_brush = QBrush(thumb_color)
        self._thumb_checked_brush = QBrush(QColor(checked_color))

        # Setup the rest of the widget.

        self.setContentsMargins(8, 0, 0, 0)
        self._thumb_position: float = 0
        self.setFixedWidth(self.FIXED_WIDTH)
        self.setFixedHeight(self.FIXED_HEIGHT)

        self.stateChanged.connect(self.thumb_state_change)

        self.thumb_anim = QPropertyAnimation(self, b"thumb_position", self)  # type: ignore
        self.thumb_anim.setEasingCurve(QEasingCurve.Type.InOutCubic)
        self.thumb_anim.setDuration(0)  # ms

        self.animations_group = QSequentialAnimationGroup()
        self.animations_group.addAnimation(self.thumb_anim)
        self._first_show = True

    def wait_for_anim_finish(self):
        # wait so all animations can finish
        start = datetime.now()
        while datetime.now() - start <= timedelta(milliseconds=self.ANIM_DURATION_MS):
            QApplication.processEvents()

    @override
    def hitButton(self, pos: QPoint):
        return self.contentsRect().contains(pos)

    @Property(float)  # type: ignore
    def thumb_position(self):  # type: ignore
        return self._thumb_position

    @thumb_position.setter
    def thumb_position(self, pos: float):
        self._thumb_position = pos
        self.update()

    @Slot(int)
    def thumb_state_change(self, value):
        self.animations_group.stop()
        if value:
            self.thumb_anim.setEndValue(1)
        else:
            self.thumb_anim.setEndValue(0)
        self.animations_group.start()

    @override
    def paintEvent(self, e: QPaintEvent):
        cont_rect = self.contentsRect()
        thumb_radius = round(self.THUMB_REL_SIZE * cont_rect.height())

        painter = QPainter(self)
        painter.setRenderHint(QPainter.RenderHint.Antialiasing)

        # Print
        painter.setPen(self._transparent_pen)
        track_rect = QRectF(
            0, 0, cont_rect.width() - thumb_radius, self.TRACK_REL_SIZE * cont_rect.height()
        )
        track_rect.moveCenter(QPointF(cont_rect.center()))
        rounding = track_rect.height() / 2

        # the thumb will move along this line
        thumb_offset: float = (
            self.TRACK_REL_SIZE - (self.THUMB_REL_SIZE * 2)
        ) * cont_rect.width()
        trail_length = cont_rect.width() + thumb_offset - 2 * thumb_radius
        x_pos = (
            cont_rect.x()
            + thumb_offset
            + thumb_radius
            + (trail_length - 3.5 * thumb_offset) * self._thumb_position
        )
        # switch brush to reflect disabled, enabled ON, and enabled OFF states
        if not self.isEnabled():
            painter.setBrush(self._disabled_checked_brush)
            painter.drawRoundedRect(track_rect, rounding, rounding)
            painter.setBrush(self._background_color_brush)
        else:
            if self.isChecked():
                painter.setBrush(self._bar_checked_brush)
                painter.drawRoundedRect(track_rect, rounding, rounding)
                painter.setBrush(self._thumb_checked_brush)
            else:
                painter.setBrush(self._background_color_brush)
                painter.drawRoundedRect(track_rect, rounding, rounding)
                painter.setPen(self._light_grey_pen)
                painter.setBrush(self._thumb_brush)

        # draw thumb
        painter.drawEllipse(QPointF(x_pos, track_rect.center().y()), thumb_radius, thumb_radius)

        painter.end()

        # set anim length after first show
        # otherwise on opening the application all toggles will slide
        if self._first_show:
            self._first_show = False
            self.thumb_anim.setDuration(self.ANIM_DURATION_MS)  # ms



from enum import Enum
from typing import Any, List, Optional, Union

from PySide6.QtCore import QModelIndex, QPersistentModelIndex, QSortFilterProxyModel, Qt
from PySide6.QtGui import QColor, QFont, QIcon
from typing_extensions import override

import conan_explorer.app as app  # using global module pattern
from conan_explorer import conan_version
from conan_explorer.app.system import get_folder_size_mb
from conan_explorer.conan_wrapper import ConanApiFactory
from conan_explorer.conan_wrapper.conan_cleanup import ConanCleanup
from conan_explorer.conan_wrapper.types import ConanPkg, ConanRef, pretty_print_pkg_info
from conan_explorer.ui.common import (
    TreeModel,
    TreeModelItem,
    get_platform_icon,
    get_themed_asset_icon,
)


class PkgSelectionType(Enum):
    ref = 0
    pkg = 1
    editable = 2
    export = 3


class PackageTreeItem(TreeModelItem):
    """Represents a tree item of a Conan pkg. To be used for the parent (ref) and the child (Profile)"""

    def __init__(
        self,
        data: List[str],
        parent=None,
        item_type=PkgSelectionType.ref,
        pkg_info={},
        invalid=False,
    ):
        super().__init__(data, parent, lazy_loading=True)
        self.pkg_info: ConanPkg = pkg_info
        self.type = item_type
        self.invalid = invalid
        if item_type == PkgSelectionType.pkg:
            self.item_data[0] = self.get_quick_profile_name()

    @override
    def load_children(self):
        # can't call super method: fetching would finish early
        self.child_items = []
        pkg_item = PackageTreeItem(["export", "0"], self, PkgSelectionType.export, ConanPkg())
        pkg_item.is_loaded = True
        self.append_child(pkg_item)
        infos = app.conan_api.get_local_pkgs_from_ref(ConanRef.loads(self.data(0)))
        for info in infos:
            pkg_item = PackageTreeItem(["package", "0"], self, PkgSelectionType.pkg, info)
            pkg_item.is_loaded = True

            self.append_child(pkg_item)
        self.is_loaded = True

    @override
    def child_count(self) -> int:
        if self.type == PkgSelectionType.ref:
            return len(self.child_items) if len(self.child_items) > 0 else 1
        return 0  # for safety

    def get_quick_profile_name(self) -> str:
        return ConanApiFactory().build_conan_profile_name_alias(
            self.pkg_info.get("settings", {})
        )


class PackageFilter(QSortFilterProxyModel):
    """Filter packages but always showing the parent (ref) of the packages"""

    def __init__(self):
        super().__init__()
        self.setFilterKeyColumn(0)

    @override
    def filterAcceptsRow(self, row_num, source_parent) -> bool:
        # Check if the current row matches
        if self.filter_accepts_row_itself(row_num, source_parent):
            return True

        # Traverse up all the way to root and check if any of them match
        if self.filter_accepts_any_parent(source_parent):
            return True

        # Finally, check if any of the children match
        return self.has_accepted_children(row_num, source_parent)

    def filter_accepts_row_itself(self, row_num, parent):
        return super(PackageFilter, self).filterAcceptsRow(row_num, parent)

    def filter_accepts_any_parent(self, parent):
        """
        Traverse to the root node and check if any of the
        ancestors match the filter
        """
        while parent.isValid():
            if self.filter_accepts_row_itself(parent.row(), parent.parent()):
                return True
            parent = parent.parent()
        return False

    def has_accepted_children(self, row_num, parent):
        """
        Starting from the current node as root, traverse all
        the descendants and test if any of the children match
        """
        model = self.sourceModel()
        source_index = model.index(row_num, 0, parent)
        source_item: Optional[PackageTreeItem] = source_index.internalPointer()
        if source_item is None:
            return False
        children_count = len(source_item.child_items)
        for i in range(children_count):
            if self.filterAcceptsRow(i, source_index):
                return True
        return False

    def lessThan(
        self,
        source_left: Union[QModelIndex, QPersistentModelIndex],
        source_right: Union[QModelIndex, QPersistentModelIndex],
    ) -> bool:
        role = Qt.ItemDataRole.DisplayRole
        left_data = self.sourceModel().data(source_left, role)
        right_data = self.sourceModel().data(source_right, role)
        if left_data is None:
            return True
        elif right_data is None:
            return False
        try:
            left_data = float(left_data)
            right_data = float(right_data)
        except Exception:
            return super().lessThan(source_left, source_right)
        if type(left_data) != type(right_data):
            # don't want to sort at all in these cases, False is just a copout ...
            # should warn user
            return False

        return left_data < right_data


class PkgSelectModel(TreeModel):
    def __init__(self, loader_signal):
        super(PkgSelectModel, self).__init__()
        # header
        self._loader_signal = loader_signal
        self._acc_size = 0
        self.root_item = PackageTreeItem(["Packages", "Size (MB)"])
        self.proxy_model = PackageFilter()
        self.proxy_model.setDynamicSortFilter(False)  # performance, should work without this
        self.proxy_model.setSourceModel(self)
        self.proxy_model.setFilterCaseSensitivity(Qt.CaseSensitivity.CaseInsensitive)
        self.proxy_model.setSortCaseSensitivity(Qt.CaseSensitivity.CaseInsensitive)
        self.show_sizes = False

    def setup_model_data(self):
        self.clear_items()
        self.beginResetModel()
        # get invalid remotes refs
        invalid_refs = []
        if conan_version.major == 1:
            invalid_refs = ConanCleanup().gather_invalid_remote_metadata()  # type: ignore
        for conan_ref in app.conan_api.get_all_local_refs():
            invalid = str(conan_ref) in invalid_refs
            conan_item = PackageTreeItem([str(conan_ref), "0"], self.root_item, invalid=invalid)
            self.root_item.append_child(conan_item)
        for conan_ref in app.conan_api.get_editable_references():
            conan_item = PackageTreeItem(
                [str(conan_ref), "0"], self.root_item, PkgSelectionType.editable
            )
            self.root_item.append_child(conan_item)
        self.endResetModel()

    def get_all_sizes(self):
        self.beginResetModel()
        for row in range(self.rowCount(QModelIndex())):
            item: PackageTreeItem = (self.index(row, 0, QModelIndex())).internalPointer()
            if item.type == PkgSelectionType.editable:
                continue
            item.load_children()  # load without expanding
            for child_item in item.child_items:
                self.get_size(child_item)
        self.endResetModel()

    def get_size(self, item: PackageTreeItem):
        if item.parent_item is None:
            return
        if item.type not in [PkgSelectionType.pkg, PkgSelectionType.export]:
            return
        if not item.item_data:
            return
        if item.item_data[1] != "0":
            return

        conan_ref = ConanRef.loads(item.parent_item.data(0))
        if item.type == PkgSelectionType.export:
            pkg_path = app.conan_api.get_export_folder(conan_ref)  # TODO can crash
        else:
            pkg_path = app.conan_api.get_package_folder(conan_ref, item.pkg_info.get("id", ""))
        self._loader_signal.emit(
            f"Calculating size for\n {str(conan_ref)}\n{self._acc_size:.1f} MB read"
        )
        size = get_folder_size_mb(pkg_path)
        item.item_data[1] = f"{size:.3f}"
        acc_size = float(item.parent_item.item_data[1]) + size
        self._acc_size += size
        item.parent_item.item_data[1] = f"{acc_size:.3f}"

    @override
    def data(self, index: Union[QModelIndex, QPersistentModelIndex], role: int = 0) -> Any:
        if not index.isValid():
            return None
        item: PackageTreeItem = index.internalPointer()  # type: ignore
        if role == Qt.ItemDataRole.ToolTipRole:
            if item.invalid:
                return "Invalid package metadata"
            if item.type == PkgSelectionType.pkg:
                # remove dict style print characters
                return pretty_print_pkg_info(item.pkg_info)

        elif role == Qt.ItemDataRole.DecorationRole:
            if index.column() != 0:
                return None
            if item.type == PkgSelectionType.ref:
                return QIcon(get_themed_asset_icon("icons/package.svg"))
            elif item.type == PkgSelectionType.editable:
                return QIcon(get_themed_asset_icon("icons/edit.svg"))
            elif item.type == PkgSelectionType.pkg:
                return get_platform_icon(item.data(index.column()))
            elif item.type == PkgSelectionType.export:
                return QIcon(get_themed_asset_icon("icons/export_notes.svg"))
        elif role == Qt.ItemDataRole.DisplayRole:
            if item.type == PkgSelectionType.editable:
                return item.data(index.column()) + " (editable)"
            else:
                return item.data(index.column())
        elif role == Qt.ItemDataRole.FontRole:
            if item.type == PkgSelectionType.editable:
                font = QFont()
                font.setItalic(True)
                return font
        # foregroundrole does not seem to work for new win11 platform so use BackgroundRole
        elif role == Qt.ItemDataRole.BackgroundRole:
            if item.invalid:
                return QColor("orange")
        return None



from pathlib import Path

from jinja2 import Template
from PySide6.QtGui import QIcon

from conan_explorer import APP_NAME, AUTHOR, BUILT_IN_PLUGIN, REPO_URL, __version__, asset_path
from conan_explorer.ui.plugin import PluginDescription, PluginInterfaceV1


class AboutPage(PluginInterfaceV1):
    def __init__(self, parent, base_signals):
        plugin_descr = PluginDescription(
            "About", BUILT_IN_PLUGIN, AUTHOR, "", "", "", " ", False, ""
        )
        super().__init__(parent, plugin_descr, base_signals=base_signals)
        from .about_ui import Ui_Form

        self._ui = Ui_Form()
        self._ui.setupUi(self)
        self.setObjectName("about_page")
        icon = QIcon(str(asset_path / "icons" / "icon.ico"))
        self._ui.logo_label.setPixmap(icon.pixmap(100, 100))

        with open(Path(__file__).parent / "about_text.html", "r") as fd:
            template = Template(fd.read())
        html_content = template.render(
            app_name=APP_NAME, version=__version__, author=AUTHOR, repo_url=REPO_URL
        )

        self._ui.about_label.setText(html_content)
